"""This is the session controller."""
import datetime as dt
import logging
from typing import Optional

from fastapi import APIRouter, Depends, HTTPException, Response
from sqlalchemy import delete, func, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload
from starlette.status import (
    HTTP_201_CREATED,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_503_SERVICE_UNAVAILABLE,
)

from ...api_models import (
    SessionCreateModel,
    SessionResult,
    SessionResultCollection,
    SessionUpdateModel,
)
from ...database.model import Node, Session, SessionNode, Tenant
from ...database.types import NodeState
from ...nodes.context import contextualize, decontextualize
from ...tasks import deprovision_nodes, fill_pools
from ..auth import req_tenant, req_tenant_optional
from ..database import req_db_async_session
from ..util import SerializationErrorRetryContext

log = logging.getLogger(__name__)

router = APIRouter(prefix="/sessions")


def wrap_with_http_422_exception(exc: Exception) -> Exception:
    return HTTPException(HTTP_422_UNPROCESSABLE_ENTITY, str(exc))


# http get http://localhost:8080/api/v1/sessions
@router.get("", response_model=SessionResultCollection, tags=["sessions"])
async def get_all_sessions(
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Optional[Tenant] = Depends(req_tenant_optional),
):
    """Return all sessions."""
    query = (
        select(Session)
        .options(
            selectinload(Session.tenant),
            selectinload(Session.session_nodes).selectinload(SessionNode.node),
        )
        .filter_by(active=True)
    )
    if tenant and not tenant.is_admin:
        query = query.filter_by(tenant=tenant)
    results = await db_async_session.execute(query)
    return {"action": "get", "sessions": results.scalars().all()}


# http get http://localhost:8080/api/v1/sessions/2
@router.get("/{id}", response_model=SessionResult, tags=["sessions"])
async def get_session(
    id: int,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Return a session with the specified **ID**."""
    session = (
        await db_async_session.execute(
            select(Session)
            .filter_by(id=id)
            .options(
                selectinload(Session.tenant),
                selectinload(Session.session_nodes).selectinload(SessionNode.node),
            )
        )
    ).scalar_one_or_none()
    if not session:
        raise HTTPException(HTTP_404_NOT_FOUND)
    if not tenant.is_admin and session.tenant != tenant:
        raise HTTPException(HTTP_403_FORBIDDEN)
    return {"action": "get", "session": session}


# http --json post http://localhost:8080/api/v1/sessions tenant_id=2 \
#     'nodes_specs:=[{"pool": "virtual-fedora34-x86_64-small", "quantity": 1}]
@router.post("", status_code=HTTP_201_CREATED, response_model=SessionResult, tags=["sessions"])
async def create_session(
    data: SessionCreateModel,
    response: Response,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    """Create a session with the requested nodes specs."""
    if tenant.is_admin and data.tenant_id is not None:
        tenant = (
            await db_async_session.execute(select(Tenant).filter_by(id=data.tenant_id))
        ).scalar_one_or_none()

        if not tenant:
            raise HTTPException(
                HTTP_422_UNPROCESSABLE_ENTITY, f"can't find tenant with id {data.tenant_id}"
            )
        elif not tenant.active:
            raise HTTPException(
                HTTP_422_UNPROCESSABLE_ENTITY, f"tenant '{tenant.name}' isn't active"
            )
    elif not tenant.is_admin and data.tenant_id is not None and data.tenant_id != tenant.id:
        raise HTTPException(HTTP_403_FORBIDDEN, "can't create session for other tenant")

    if not tenant.is_admin:
        requested_nodes = sum(nodes_spec.quantity for nodes_spec in data.nodes_specs)

        current_allocation = (
            await db_async_session.execute(
                select(func.count())
                .select_from(SessionNode)
                .join(Session, Session.id == SessionNode.session_id)
                .join(Node, Node.id == SessionNode.node_id)
                .filter(Session.active == True, Session.tenant == tenant)  # noqa: E712
            )
        ).scalar_one()

        if requested_nodes + current_allocation > tenant.effective_node_quota:
            raise HTTPException(
                HTTP_403_FORBIDDEN,
                f"quota exceeded: requested nodes ({requested_nodes}) + current allocation"
                f" ({current_allocation}) = {requested_nodes + current_allocation}"
                f" > {tenant.effective_node_quota}",
            )

    # Detach the tenant from the session => only access loaded attributes below.
    db_async_session.expunge(tenant)
    await db_async_session.commit()

    log.debug("Attempting to allocate nodes for: %s", data.nodes_specs)

    async with SerializationErrorRetryContext(
        exception_wrapper=wrap_with_http_422_exception
    ) as retry:
        async for attempt in retry.attempts:
            try:
                async with db_async_session.begin():
                    session = Session(
                        tenant_id=tenant.id,
                        data={"nodes_specs": [spec.model_dump() for spec in data.nodes_specs]},
                        expires_at=(
                            dt.datetime.utcnow().replace(tzinfo=dt.timezone.utc)
                            + tenant.effective_session_lifetime
                        ),
                    )
                    db_async_session.add(session)
                    await db_async_session.flush()

                    session_id = session.id

                    session_nodes = []
                    pools_to_fill_up = set()
                    for nodes_spec in data.nodes_specs:
                        pools_to_fill_up.add(nodes_spec.pool)
                        nodes_spec_dict = nodes_spec.model_dump()
                        quantity = nodes_spec_dict.pop("quantity")

                        query = (
                            select(Node)
                            .filter_by(active=True, state=NodeState.ready, **nodes_spec_dict)
                            .limit(quantity)
                            .with_for_update()
                        )

                        nodes_to_reserve = (await db_async_session.execute(query)).scalars().all()

                        if len(nodes_to_reserve) < quantity:
                            raise HTTPException(
                                HTTP_422_UNPROCESSABLE_ENTITY, f"can't reserve nodes: {nodes_spec}"
                            )

                        # take the nodes out of circulation and update data
                        for node in nodes_to_reserve:
                            # record why this node was allocated for this session
                            node.data["nodes_spec"] = nodes_spec.model_dump()
                            node.state = NodeState.contextualizing
                            session_node = SessionNode(
                                session=session, node=node, pool=nodes_spec.pool, data=node.data
                            )
                            session_nodes.append(session_node)
                            db_async_session.add(session_node)
                            log.debug("Allocating node: %s (%s)", node.id, node.pool)
            except retry.exceptions as exc:
                retry.process_exception(exc)

    # Unfortunately, it is possible that the read operations on nodes above (in another concurrent
    # request) conflict with the write operations on (as far as the respective result sets go,
    # unrelated) nodes below. Therefore, retry this a couple of times before giving up.

    async with SerializationErrorRetryContext(
        exception_wrapper=wrap_with_http_422_exception
    ) as retry:
        async for attempt in retry.attempts:
            try:
                async with db_async_session.begin():
                    # New transaction -> reload session and related node objects

                    session = (
                        await db_async_session.execute(
                            select(Session)
                            .filter_by(id=session_id)
                            .options(
                                selectinload(Session.tenant),
                                selectinload(Session.session_nodes).selectinload(SessionNode.node),
                            )
                        )
                    ).scalar_one()

                    nodes_in_transaction = sorted(
                        (sn.node for sn in session.session_nodes), key=lambda node: node.id
                    )

                    # Mark the nodes as deployed but only commit when they’re contextualized.
                    for node in nodes_in_transaction:
                        node.state = NodeState.deployed
                    await db_async_session.flush()

                    contextualized_ipaddrs = await contextualize(
                        nodes=[node.ipaddr for node in nodes_in_transaction],
                        ssh_pubkey=tenant.ssh_key,
                    )

                    if None in contextualized_ipaddrs:
                        try:
                            # Undo the session and related objects being added to the database
                            # above.
                            await db_async_session.execute(
                                delete(SessionNode).filter_by(session_id=session.id)
                            )
                            await db_async_session.execute(delete(Session).filter_by(id=session.id))

                            log.error("One or more nodes couldn't be contextualized:")
                            nodes_to_decontextualize = []
                            for node, ipaddr in zip(nodes_in_transaction, contextualized_ipaddrs):
                                if not ipaddr:
                                    log.error(
                                        "    id: %s hostname: %s ipaddr: %s",
                                        node.id,
                                        node.hostname,
                                        node.ipaddr,
                                    )
                                    node.fail("contextualizing node failed")
                                else:
                                    nodes_to_decontextualize.append(node)

                            decontextualized_ipaddrs = await decontextualize(
                                nodes=[node.ipaddr for node in nodes_to_decontextualize]
                            )

                            if None in decontextualized_ipaddrs:
                                log.error("One or more nodes couldn't be decontextualized:")
                                for node, ipaddr in zip(
                                    nodes_to_decontextualize, decontextualized_ipaddrs
                                ):
                                    if not ipaddr:
                                        log.error(
                                            "    id: %s hostname: %s ipaddr: %s",
                                            node.id,
                                            node.hostname,
                                            node.ipaddr,
                                        )
                                        node.fail("decontextualizing node failed")
                                    else:
                                        node.state = NodeState.ready

                            await db_async_session.commit()
                        except Exception as exc:
                            try:
                                await db_async_session.commit()
                            except Exception:  # pragma: no cover
                                pass
                            raise HTTPException(
                                HTTP_503_SERVICE_UNAVAILABLE,
                                "decontextualizing nodes failed after contextualization failure",
                                headers={"Retry-After": "0"},
                            ) from exc

                        # Some nodes are out of circulation, fill up pools.
                        fill_pools.delay(pool_names=list(pools_to_fill_up)).forget()

                        raise HTTPException(
                            HTTP_503_SERVICE_UNAVAILABLE,
                            "contextualization of nodes failed",
                            headers={"Retry-After": "0"},
                        )
                    # None not in contextualized_ipaddrs
                    # Workaround: give coverage an anchor to detect that this branch is taken.
                    pass
            except retry.exceptions as exc:
                retry.process_exception(exc)

    log.debug("Nodes deployed, kick off filling pools and return result via API")

    # Tell backend worker to fill up pools from which nodes were taken.
    fill_pools.delay(pool_names=list(pools_to_fill_up)).forget()

    return {"action": "post", "session": session}


# http --json put http://localhost:8080/api/v1/sessions/2 active:=false
@router.put("/{id}", response_model=SessionResult, tags=["sessions"])
async def update_session(
    id: int,
    data: SessionUpdateModel,
    db_async_session: AsyncSession = Depends(req_db_async_session),
    tenant: Tenant = Depends(req_tenant),
):
    session = (
        await db_async_session.execute(
            select(Session)
            .filter_by(id=id)
            .options(
                selectinload(Session.tenant),
                selectinload(Session.session_nodes).selectinload(SessionNode.node),
            )
        )
    ).scalar_one_or_none()

    if data.expires_at is not None:
        if isinstance(data.expires_at, dt.timedelta):
            new_expires_at = session.expires_at + data.expires_at
        else:  # isinstance(data.expires_at, dt.datetime)
            new_expires_at = data.expires_at.replace(tzinfo=dt.timezone.utc)

        # Clamp to allowable values
        new_expires_at = max(new_expires_at, session.created_at)

        if not tenant.is_admin:
            new_expires_at = min(
                new_expires_at, session.created_at + tenant.effective_session_lifetime_max
            )

        session.expires_at = new_expires_at

    if not session:
        raise HTTPException(HTTP_404_NOT_FOUND)

    if not tenant.is_admin and session.tenant != tenant:
        raise HTTPException(HTTP_403_FORBIDDEN)

    if not session.active:
        raise HTTPException(HTTP_422_UNPROCESSABLE_ENTITY, f"session {id} is retired")

    if data.active is False:
        session.active = data.active
        deprovision_nodes.delay(
            node_ids=[session_node.node_id for session_node in session.session_nodes]
        ).forget()

    await db_async_session.flush()

    return {"action": "put", "session": session}



from abc import ABC
from datetime import timedelta
from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, SecretStr, model_validator
from typing_extensions import Annotated

from ..misc import ConfigTimeDelta
from .common import APIResult, CreatableMixin, RetirableMixin

# tenant model


class TenantBase(BaseModel, ABC):
    name: str
    is_admin: Optional[bool] = None
    ssh_key: SecretStr
    node_quota: Optional[Annotated[int, Field(gt=0)]] = None
    session_lifetime: Optional[ConfigTimeDelta] = None
    session_lifetime_max: Optional[ConfigTimeDelta] = None
    model_config = ConfigDict(from_attributes=True, ser_json_timedelta="float")


class TenantCreateModel(TenantBase):
    pass


class TenantRetireModel(BaseModel):
    active: bool


class TenantUpdateModel(BaseModel):
    ssh_key: Optional[SecretStr] = None
    api_key: Optional[Union[UUID, Literal["reset"]]] = None
    node_quota: Optional[Annotated[int, Field(gt=0)]] = None
    session_lifetime: Optional[ConfigTimeDelta] = None
    session_lifetime_max: Optional[ConfigTimeDelta] = None
    model_config = ConfigDict(
        minimum_fields=(
            "ssh_key",
            "api_key",
            "node_quota",
            "session_lifetime",
            "session_lifetime_max",
        )
    )

    @model_validator(mode="before")
    @classmethod
    def check_any_field_set(cls, values):
        if not any(field in values for field in cls.model_config["minimum_fields"]):
            raise ValueError(f"one of {', '.join(cls.model_config['minimum_fields'])} is required")
        return values


class TenantModel(TenantBase, CreatableMixin, RetirableMixin):
    id: int
    effective_node_quota: int
    effective_session_lifetime: timedelta
    effective_session_lifetime_max: timedelta


class TenantCreateResultModel(TenantModel):
    ssh_key: str
    api_key: UUID


class TenantUpdateResultModel(TenantModel):
    api_key: Optional[Union[UUID, SecretStr]] = None


# API results


class TenantResult(APIResult):
    tenant: TenantModel


class TenantCreateResult(TenantResult):
    tenant: TenantCreateResultModel


class TenantUpdateResult(TenantResult):
    tenant: TenantUpdateResultModel


class TenantResultCollection(APIResult):
    tenants: List[TenantModel]



import json
import shlex
from typing import Generator

import yaml

from .main import DuffyAPIErrorModel, JSONValue


class DuffyFormatter:
    _subclasses_for_format = {}

    def __init_subclass__(cls, format, **kwargs):
        cls._subclasses_for_format[format] = cls

    @classmethod
    def new_for_format(cls, format, *args, **kwargs):
        return cls._subclasses_for_format[format](*args, **kwargs)

    def format(self, result: JSONValue) -> str:
        raise NotImplementedError()


class DuffyJSONFormatter(DuffyFormatter, format="json"):
    def format(self, result: JSONValue) -> str:
        return json.dumps(result)


class DuffyYAMLFormatter(DuffyFormatter, format="yaml"):
    def format(self, result: JSONValue) -> str:
        return yaml.dump(result)


class DuffyFlatFormatter(DuffyFormatter, format="flat"):
    field_name_to_flattener = {
        "error": "flatten_api_error",
        "pool": "flatten_pool_result",
        "pools": "flatten_pools_result",
        "session": "flatten_session_result",
        "sessions": "flatten_sessions_result",
    }

    @staticmethod
    def format_key_value(key, value):
        if value is None:
            value = ""
        elif isinstance(value, bool):
            value = "TRUE" if value else "FALSE"
        elif isinstance(value, (int, float)):
            pass
        else:
            value = shlex.quote(str(value))
            if value[:1] != "'":
                value = f"'{value}'"

        return f"{key}={value}"

    def flatten_api_error(self, api_error: DuffyAPIErrorModel) -> Generator[str, None, None]:
        yield self.format_key_value("error", api_error.error.detail)

    def flatten_pool(self, pool: JSONValue) -> Generator[str, None, None]:
        fields = {
            "pool_name": pool["name"],
            "fill_level": pool.get("fill-level", pool.get("fill_level")),
        }
        if "levels" in pool:
            fields.update(
                {
                    "levels_provisioning": pool["levels"]["provisioning"],
                    "levels_ready": pool["levels"]["ready"],
                    "levels_contextualizing": pool["levels"]["contextualizing"],
                    "levels_deployed": pool["levels"]["deployed"],
                    "levels_deprovisioning": pool["levels"]["deprovisioning"],
                }
            )
        yield " ".join(self.format_key_value(key, value) for key, value in fields.items())

    def flatten_pool_result(self, result: JSONValue) -> Generator[str, None, None]:
        yield from self.flatten_pool(result["pool"])

    def flatten_pools_result(self, result: JSONValue) -> Generator[str, None, None]:
        for pool in result["pools"]:
            yield from self.flatten_pool(pool)

    def flatten_session(self, session: JSONValue) -> Generator[str, None, None]:
        for node in sorted(
            session["nodes"], key=lambda node: (node["pool"], node["hostname"], node["ipaddr"])
        ):
            fields = {
                "session_id": session["id"],
                "active": session["active"],
                "created_at": session["created_at"],
                "retired_at": session["retired_at"],
                "pool": node["pool"],
                "hostname": node["hostname"],
                "ipaddr": node["ipaddr"],
            }
            yield " ".join(self.format_key_value(key, value) for key, value in fields.items())

    def flatten_session_result(self, result: JSONValue) -> Generator[str, None, None]:
        yield from self.flatten_session(result["session"])

    def flatten_sessions_result(self, result: JSONValue) -> Generator[str, None, None]:
        for session in result["sessions"]:
            yield from self.flatten_session(session)

    def format(self, result: JSONValue) -> str:
        for field_name, flattener in self.field_name_to_flattener.items():
            if field_name in result:
                return "\n".join(getattr(self, flattener)(result))

        raise TypeError("Can't flatten {result!r}")



import logging
from pathlib import Path
from typing import Set

import alembic.command
import alembic.config
import alembic.runtime.environment
import alembic.script

from ...configuration import config

log = logging.getLogger(__name__)

HERE = Path(__file__).parent


class AlembicMigration:
    """Glue Duffy (mainly its config) and Alembic DB migrations together."""

    @property
    def config(self):
        if not hasattr(self, "_config"):
            self._config = alembic.config.Config()
            self._config.set_main_option("script_location", str(HERE.absolute()).replace("%", "%%"))
            self._config.set_main_option(
                "sqlalchemy.url", config["database"]["sqlalchemy"]["sync_url"].replace("%", "%%")
            )
        return self._config

    def create(self, comment: str, autogenerate: bool):
        alembic.command.revision(config=self.config, message=comment, autogenerate=autogenerate)

        if autogenerate:
            log.warning(
                ">>> Remember to edit the autogenerated migration script! Unedited, it would drop"
                " any support table not registered with your SQLAlchemy metadata."
            )

    def _get_current(self) -> Set[str]:
        script = alembic.script.ScriptDirectory.from_config(self.config)

        current_revs = set()

        def _get_rev_current(rev, context):
            current_revs.update(
                _rev.cmd_format(verbose=False) for _rev in script.get_all_current(rev)
            )
            return []

        with alembic.runtime.environment.EnvironmentContext(
            self.config, script, fn=_get_rev_current, dont_mutate=True
        ):
            script.run_env()

        return current_revs

    def db_version(self):
        print("\n".join(self._get_current()))

    def upgrade(self, version: str):
        pre_revs = self._get_current()
        alembic.command.upgrade(self.config, version)
        post_revs = self._get_current()
        if pre_revs == post_revs:
            print("Nothing to upgrade.")
        else:
            print("Upgraded to:", ", ".join(post_revs))

    def downgrade(self, version: str):
        pre_revs = self._get_current()
        alembic.command.downgrade(self.config, version)
        post_revs = self._get_current()
        if pre_revs == post_revs:
            print("Nothing to downgrade.")
        else:
            print("Downgraded to:", ", ".join(post_revs) if post_revs else "<base>")


alembic_migration = AlembicMigration()



"""Add Node.state, SessionNode.session_id indexes

Revision ID: 6654d536b836
Revises: 36cf064c0481
Create Date: 2023-09-06 21:42:56.647045
"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "6654d536b836"
down_revision = "36cf064c0481"
branch_labels = None
depends_on = None


def upgrade():
    op.create_index(op.f("nodes_state_index"), "nodes", ["state"], unique=False)
    op.create_index(
        op.f("sessions_nodes_session_id_index"), "sessions_nodes", ["session_id"], unique=False
    )


def downgrade():
    op.drop_index(op.f("sessions_nodes_session_id_index"), table_name="sessions_nodes")
    op.drop_index(op.f("nodes_state_index"), table_name="nodes")



from abc import ABC
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Annotated

try:
    from ..database.types import NodeState
except ImportError:  # pragma: no cover
    NodeState = str
from ..misc import APITimeDelta
from .common import APIResult, CreatableMixin, RetirableMixin
from .node import NodeBase
from .tenant import TenantModel

# nodes spec model


class NodesSpec(BaseModel):
    quantity: Annotated[int, Field(ge=1)]
    pool: str


# nodes in sessions model


class SessionNodeModel(NodeBase):
    id: int
    state: Optional[NodeState] = None


# session model


class SessionBase(BaseModel, ABC):
    model_config = ConfigDict(from_attributes=True)


class SessionCreateModel(SessionBase):
    tenant_id: Optional[int] = None
    nodes_specs: List[NodesSpec]


class SessionUpdateModel(SessionBase):
    active: Optional[bool] = None
    expires_at: Optional[Union[datetime, APITimeDelta]] = None


class SessionModel(SessionBase, CreatableMixin, RetirableMixin):
    id: int
    expires_at: Optional[datetime] = None
    tenant: TenantModel
    data: Dict[str, Any]
    nodes: List[SessionNodeModel]


# API results


class SessionResult(APIResult):
    session: SessionModel


class SessionResultCollection(APIResult):
    sessions: List[SessionModel]



from enum import Enum
from http import HTTPStatus
from typing import Any, Dict, List, Optional, Sequence, Union

import httpx
from pydantic import BaseModel, ConfigDict

from ..api_models import SessionCreateModel, SessionUpdateModel
from ..configuration import config

JSONValue = Union[None, bool, str, float, int, List["JSONValue"], Dict[str, "JSONValue"]]


class _MethodEnum(str, Enum):
    get = "get"
    post = "post"
    put = "put"
    delete = "delete"


class DuffyApiErrorDetailModel(BaseModel):
    detail: str
    model_config = ConfigDict(extra="allow")


class DuffyAPIErrorModel(BaseModel):
    error: DuffyApiErrorDetailModel


class DuffyClient:
    def __init__(
        self,
        url: Optional[str] = None,
        auth_name: Optional[str] = None,
        auth_key: Optional[str] = None,
    ):
        if url:
            self.url = url
        if auth_name:
            self.auth_name = auth_name
        if auth_key:
            self.auth_key = auth_key

    @property
    def url(self):
        return getattr(self, "_url", None) or config["client"]["url"]

    @url.setter
    def url(self, value):
        self._url = value

    @property
    def auth_name(self):
        return getattr(self, "_auth_name", None) or config["client"]["auth"]["name"]

    @auth_name.setter
    def auth_name(self, value):
        self._auth_name = value

    @property
    def auth_key(self):
        return getattr(self, "_auth_key", None) or config["client"]["auth"]["key"]

    @auth_key.setter
    def auth_key(self, value):
        self._auth_key = value

    def client(self):
        return httpx.Client(auth=(self.auth_name, self.auth_key), base_url=self.url, timeout=None)

    def _query_method(
        self,
        method: _MethodEnum,
        url: str,
        *,
        in_dict: Optional[Dict[str, Any]] = None,
        in_model: Optional[BaseModel] = None,
        expected_status: Union[HTTPStatus, Sequence[HTTPStatus]] = HTTPStatus.OK,
    ) -> JSONValue:
        add_kwargs = {}
        if in_dict is not None:
            add_kwargs["json"] = in_model(**in_dict).model_dump()

        with self.client() as client:
            client_method = getattr(client, method.name)
            response = client_method(url=url, **add_kwargs)

        if isinstance(expected_status, HTTPStatus):
            expected_status = (expected_status,)

        if response.status_code not in expected_status:
            try:
                return DuffyAPIErrorModel(error=response.json()).model_dump(by_alias=True)
            except Exception as exc:
                response.raise_for_status()
                raise RuntimeError(f"Can't process response: {response}") from exc

        return response.json()

    def list_sessions(self) -> JSONValue:
        return self._query_method(_MethodEnum.get, "/sessions")

    def show_session(self, session_id: int) -> JSONValue:
        return self._query_method(_MethodEnum.get, f"/sessions/{session_id}")

    def request_session(self, nodes_specs: List[Dict[str, str]]) -> JSONValue:
        return self._query_method(
            _MethodEnum.post,
            "/sessions",
            in_dict={"nodes_specs": nodes_specs},
            in_model=SessionCreateModel,
            expected_status=HTTPStatus.CREATED,
        )

    def retire_session(self, session_id: int) -> JSONValue:
        return self._query_method(
            _MethodEnum.put,
            f"/sessions/{session_id}",
            in_dict={"active": False},
            in_model=SessionUpdateModel,
        )

    def list_pools(self) -> JSONValue:
        return self._query_method(_MethodEnum.get, "/pools")

    def show_pool(self, pool_name: str) -> JSONValue:
        return self._query_method(_MethodEnum.get, f"/pools/{pool_name}")



import datetime as dt
import re
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Union

from pydantic import (
    GetCoreSchemaHandler,
    PlainSerializer,
    TypeAdapter,
    WithJsonSchema,
    WrapValidator,
)
from typing_extensions import Annotated, _AnnotatedAlias


@dataclass(frozen=True)
class CustomTimeDelta:
    strip_whitespace: bool = True
    to_lower: bool = True
    needs_sign: bool = False
    allow_negative: bool = True
    allow_dimensionless_seconds: bool = True
    serialize_to_seconds: bool = False

    @cached_property
    def timedelta_re(self) -> re.Pattern:
        sign = "+-" if self.allow_negative else "+"
        needs_sign = "" if self.needs_sign else "?"
        return re.compile(
            rf"(?P<sign>[{sign}]){needs_sign}(?!\s*$)"
            + r"(?:\s*(?P<weeks>\d+)w)?"
            + r"(?:\s*(?P<days>\d+)d)?"
            + r"(?:\s*(?P<hours>\d+)h)?"
            + r"(?:\s*(?P<minutes>\d+)m)?"
            + r"(?:\s*(?P<seconds>\d+)s)?"
            + r"(?:\s*(?P<milliseconds>\d+)ms)?"
        )

    def validate(self, v: Union[str, Any], handler: GetCoreSchemaHandler) -> dt.timedelta:
        if isinstance(v, dt.timedelta):
            return handler(v)

        if self.allow_dimensionless_seconds:
            if isinstance(v, (int, float)):
                return dt.timedelta(seconds=v)
            if not isinstance(v, str):
                raise TypeError(f"input value {v!r} must be a string, an integer or a float")
            if v.isdigit():
                return dt.timedelta(seconds=int(v))
        elif not isinstance(v, str):
            raise TypeError("input value must be a string")

        if self.strip_whitespace:
            v = v.strip()

        if self.to_lower:
            v = v.lower()

        m = self.timedelta_re.fullmatch(v)
        if not m:
            raise ValueError("invalid timedelta format")

        sign = -1 if m.group("sign") == "-" else 1
        return dt.timedelta(
            weeks=sign * int(m.group("weeks") or 0),
            days=sign * int(m.group("days") or 0),
            hours=sign * int(m.group("hours") or 0),
            minutes=sign * int(m.group("minutes") or 0),
            seconds=sign * int(m.group("seconds") or 0),
            milliseconds=sign * int(m.group("milliseconds") or 0),
        )

    def serialize(self, v: dt.timedelta) -> Union[str, int]:
        total_seconds = v.total_seconds()

        if self.serialize_to_seconds:
            if not v.microseconds:
                return int(total_seconds)
            else:
                return total_seconds

        chunks: dict[str, int] = {}

        if total_seconds < 0:
            sign = "-"
            v = -v
        elif self.needs_sign:
            sign = "+"
        else:
            sign = ""

        weeks = v.days // 7
        days = v.days % 7
        if weeks:
            chunks["w"] = weeks
        if days:
            chunks["d"] = days

        hours = v.seconds // 3600
        hour_seconds = v.seconds % 3600
        minutes = hour_seconds // 60
        seconds = hour_seconds % 60

        if hours:
            chunks["h"] = hours
        if minutes:
            chunks["m"] = minutes
        if seconds:
            chunks["s"] = seconds

        milliseconds = v.microseconds // 1000
        if milliseconds:
            chunks["ms"] = milliseconds

        if not chunks:
            chunks["s"] = 0

        return sign + "".join(f"{value}{key}" for key, value in chunks.items())

    @cached_property
    def pydantic_type(self) -> _AnnotatedAlias:
        json_schema = TypeAdapter(dt.timedelta).json_schema()
        json_schema.update(examples=["+3h30m"])

        if self.needs_sign:
            json_schema.update(pattern=r"^[-+](\d+w)?(\d+d)?(\d+h)?(\d+m)?(\d+s)?(\d+ms)?$")
        else:
            json_schema.update(pattern=r"^[-+]?(\d+w)?(\d+d)?(\d+h)?(\d+m)?(\d+s)?(\d+ms)?$")
            json_schema["examples"].append("1h")

        if self.allow_negative:
            json_schema["examples"].append("-20m25s")

        return Annotated[
            dt.timedelta,
            WrapValidator(self.validate),
            PlainSerializer(self.serialize),
            WithJsonSchema(json_schema),
        ]


ConfigTimeDelta = CustomTimeDelta(allow_negative=False, serialize_to_seconds=True).pydantic_type
APITimeDelta = CustomTimeDelta(needs_sign=True, allow_dimensionless_seconds=False).pydantic_type



import asyncio
from copy import deepcopy

from sqlalchemy import MetaData, create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.engine.interfaces import DBAPIConnection
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
from sqlalchemy.orm import declarative_base, sessionmaker
from sqlalchemy.pool import ConnectionPoolEntry

from ..configuration import config
from ..exceptions import DuffyConfigurationError

# use custom metadata to specify naming convention
naming_convention = {
    "ix": "%(column_0_N_label)s_index",
    "uq": "%(table_name)s_%(column_0_N_name)s_key",
    "ck": "%(table_name)s_%(constraint_name)s_check",
    "fk": "%(table_name)s_%(column_0_N_name)s_%(referred_table_name)s_fkey",
    "pk": "%(table_name)s_pkey",
}
metadata = MetaData(naming_convention=naming_convention)
Base = declarative_base(metadata=metadata)

async_session_maker = sessionmaker(class_=AsyncSession, expire_on_commit=False, future=True)
sync_session_maker = sessionmaker(future=True, expire_on_commit=False)


def init_sync_model(sync_engine: Engine = None):
    if not sync_engine:
        sync_engine = get_sync_engine()
    sync_session_maker.configure(bind=sync_engine)


async def init_async_model(async_engine: AsyncEngine = None):
    if not async_engine:
        async_engine = get_async_engine()
    async_session_maker.configure(bind=async_engine)


def init_model(sync_engine: Engine = None, async_engine: AsyncEngine = None):
    init_sync_model(sync_engine)
    asyncio.run(init_async_model(async_engine))


_key_failed_to_config_key = {
    "sqlalchemy": "database.sqlalchemy",
    "sync_url": "database.sqlalchemy.sync_url",
    "async_url": "database.sqlalchemy.async_url",
}


def _pgsql_disable_seqscan(
    dbapi_connection: DBAPIConnection, connection_record: ConnectionPoolEntry
) -> None:
    """Disables the query planner's use of sequential scan plan types.

    As far as this is possible at least.

    Sequential scanning in queries can cause conflicts in concurrent transactions even if the
    respective rows accessed in the transactions are different, merely iterating over the rows “of
    the other transaction” can cause them to be locked.

    For this to work, it is necessary that the involved columns have an index, i.e. the planner has
    a usable alternative to sequential scans."""
    cursor = dbapi_connection.cursor()
    cursor.execute("SET enable_seqscan=off")
    cursor.close()


def get_sync_engine():
    try:
        sync_config = deepcopy(config["database"]["sqlalchemy"]) or {}
        sync_config["url"] = sync_config.pop("sync_url")
    except (AttributeError, KeyError) as exc:
        key_not_found = exc.args[0]
        raise DuffyConfigurationError(
            _key_failed_to_config_key.get(key_not_found, key_not_found)
        ) from exc
    sync_config.pop("async_url", None)
    sync_config.setdefault("isolation_level", "SERIALIZABLE")
    engine = create_engine(**sync_config)
    if engine.name == "postgresql":
        event.listen(engine, "connect", _pgsql_disable_seqscan)
    return engine


def get_async_engine():
    try:
        async_config = deepcopy(config["database"]["sqlalchemy"]) or {}
        async_config["url"] = async_config.pop("async_url")
    except (AttributeError, KeyError) as exc:
        key_not_found = exc.args[0]
        raise DuffyConfigurationError(
            _key_failed_to_config_key.get(key_not_found, key_not_found)
        ) from exc
    async_config.pop("sync_url", None)
    async_config.setdefault("isolation_level", "SERIALIZABLE")
    engine = create_async_engine(**async_config)
    if engine.name == "postgresql":
        event.listen(engine.sync_engine, "connect", _pgsql_disable_seqscan)
    return engine



from typing import Tuple

from celery import Celery

from .. import database
from ..configuration import config
from ..configuration.validation import PeriodicTaskModel
from ..nodes.pools import NodePool
from .base import celery, init_tasks
from .expire import expire_sessions
from .provision import fill_pools

DEFAULT_PERIODIC_INTERVAL = 5 * 60


@celery.on_after_finalize.connect
def setup_periodic_tasks(sender: Celery, **kwargs):
    periodic_config = config.get("tasks", {}).get("periodic", {})

    fill_pools_interval = PeriodicTaskModel(
        **periodic_config.get("fill-pools", {"interval": DEFAULT_PERIODIC_INTERVAL})
    ).interval

    expire_sessions_interval = PeriodicTaskModel(
        **periodic_config.get("expire-sessions", {"interval": DEFAULT_PERIODIC_INTERVAL})
    ).interval

    sender.add_periodic_task(fill_pools_interval.total_seconds(), fill_pools.signature())
    sender.add_periodic_task(expire_sessions_interval.total_seconds(), expire_sessions.signature())


@celery.on_after_finalize.connect
def run_init_tasks(sender: Celery, **kwargs):
    fill_pools.delay().forget()
    expire_sessions.delay().forget()


def start_worker(worker_args: Tuple[str]):
    database.init_sync_model()
    init_tasks()
    NodePool.process_configuration()

    celery.worker_main(("worker",) + worker_args)



import uuid
from collections import UserString
from logging import Filter, LogRecord
from typing import Any, Dict

from .middleware import request_id_ctxvar


class LazyShortRequestId(UserString):
    def __init__(self, request_id: uuid.UUID) -> None:
        self._request_id = request_id

    def __str__(self) -> str:
        if not hasattr(self, "_short"):
            self._short = str(self._request_id)[-12:]
        return self._short

    @property
    def data(self):
        return self.__str__()


class LazyFormattedString(UserString):
    def __init__(self, fmt: str, **kwargs: Dict[str, Any]) -> None:
        self._fmt = fmt
        self._kwargs = kwargs

    def __str__(self) -> str:
        if not hasattr(self, "_formatted"):
            self._formatted = self._fmt.format(**self._kwargs)
        return self._formatted

    @property
    def data(self):
        return self.__str__()


class RequestIdFilter(Filter):
    def filter(self, record: LogRecord) -> bool:
        record.request_id = rid = request_id_ctxvar.get()

        if rid:
            # Use lazily evaluated objects here, so they’re only formatted when actually accessed.
            record.short_request_id = srid = LazyShortRequestId(rid)

            # Note the trailing spaces
            record.request_id_optional = LazyFormattedString("[{rid}] ", rid=rid)
            record.short_request_id_optional = LazyFormattedString("[{srid}] ", srid=srid)
        else:
            record.short_request_id = None
            record.request_id_optional = record.short_request_id_optional = ""

        return True



import asyncio
import enum
import logging
import time
from random import random
from typing import AsyncIterator, Callable, Iterator, Optional, Tuple, Union

log = logging.getLogger(__name__)


def camel_case_to_lower_with_underscores(camelcased: str) -> str:
    """Convert CamelCased names to lower_case_with_underscores."""
    chunk_positions = []
    prev_split_pos = 0

    for pos, (prev_is_upper, char_is_upper, next_is_upper) in enumerate(
        zip(
            (x.isupper() for x in camelcased[:-2]),
            (x.isupper() for x in camelcased[1:-1]),
            (x.isupper() for x in camelcased[2:]),
        ),
        start=1,
    ):
        if char_is_upper and (not prev_is_upper or not next_is_upper):
            chunk_positions.append((prev_split_pos, pos))
            prev_split_pos = pos

    chunk_positions.append((prev_split_pos, len(camelcased)))

    lowercased = camelcased.lower()

    return "_".join(
        lowercased[prev_split_pos:split_pos] for prev_split_pos, split_pos in chunk_positions
    )


def merge_dicts(*src_dicts):
    """Create a deep merge of several dictionaries.

    The structure of the dictionaries must be compatible, i.e. sub-dicts
    may not be differently typed between the source dictionaries."""
    if not src_dicts:
        raise ValueError("Can't merge nothing")

    if not all(isinstance(src_dict, dict) for src_dict in src_dicts):
        raise TypeError("All objects to be merged have to be dictionaries")

    res_dict = {}

    for src_dict in src_dicts:
        for key, src_value in src_dict.items():
            if isinstance(src_value, dict):
                if key not in res_dict:
                    res_dict[key] = src_value.copy()
                else:
                    res_dict[key] = merge_dicts(res_dict[key], src_value)
            elif key in res_dict and isinstance(res_dict[key], dict):
                raise TypeError("All objects to be merged have to be dictionaries")
            else:
                res_dict[key] = src_value

    return res_dict


class SentinelType(enum.Enum):
    UNSET = 1


UNSET = SentinelType.UNSET
"""A sentinel object for unset values."""


IntOrFloat = Union[int, float]


class RetryContext:
    """A wrapper for code blocks that should be retried on certain exceptions.

    Use it e.g. like this:

        async with RetryContext(exceptions=RuntimeError) as retry:
            async for attempt in retry.attempts:
                # ... set up the things ...
                try:
                    # ... do the things ...
                except retry.exceptions as exc:
                    # ... undo the things ...
                    retry.process_exception(exc)
                finally:
                    # ... tear down the things ...

    Subclass it e.g. to inspect caught exceptions further in
    process_exception(), making a decision whether to re-raise or continue
    with attempts.
    """

    exceptions: Union[Exception, Tuple[Exception]] = Exception
    no_attempts: int = 5
    delay_min: IntOrFloat = 0.1
    delay_max: IntOrFloat = 1.6
    delay_backoff_factor: IntOrFloat = 2
    delay_add_fuzz: IntOrFloat = 0.3
    exception_wrapper: Optional[Callable[[Exception], Exception]] = None

    def __init__(
        self,
        *,
        exceptions: Union[Exception, Tuple[Exception]] = None,
        no_attempts: int = None,
        delay_min: IntOrFloat = None,
        delay_max: IntOrFloat = None,
        delay_backoff_factor: IntOrFloat = None,
        delay_add_fuzz: IntOrFloat = None,
        exception_wrapper: Optional[Callable[[Exception], Exception]] = None,
    ):
        self._is_async = None

        if exceptions is not None:
            self.exceptions = exceptions

        if no_attempts is not None:
            self.no_attempts = no_attempts

        if delay_min is not None:
            self.delay_min = delay_min

        if delay_max is not None:
            self.delay_max = delay_max

        if delay_backoff_factor is not None:
            self.delay_backoff_factor = delay_backoff_factor

        if delay_add_fuzz is not None:
            self.delay_add_fuzz = delay_add_fuzz

        if exception_wrapper is not None:
            self.exception_wrapper = exception_wrapper

    def __repr__(self):
        pieces = [
            f"{name}={value!r}"
            for name in ("exceptions", "no_attempts")
            for value in (getattr(self, name),)
            if value is not None
        ]
        return f"{type(self).__name__}({', '.join(pieces)})"

    def __enter__(self):
        self._is_async = False
        self._exc = None
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    async def __aenter__(self):
        self._is_async = True
        self._exc = None
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        pass

    def wrap_exception(self, exc: Exception) -> Exception:
        if self.exception_wrapper:
            exc = self.exception_wrapper(exc)
        return exc

    async def _async_attempts(self) -> AsyncIterator[int]:
        delay = self.delay_min

        for attempt in range(1, self.no_attempts + 1):
            log.debug("[%r] Attempt %d of %d", self, attempt, self.no_attempts)

            self._exc = None

            yield attempt

            if not self._exc:
                break

            if attempt < self.no_attempts:
                log.debug("[%r] Retrying...", self)
                current_delay = delay + self.delay_add_fuzz * random()
                await asyncio.sleep(current_delay)
                delay = max(self.delay_max, delay * self.delay_backoff_factor)

        if self._exc:
            log.warning(
                "[%r] Number of attempts (%d) exhausted, re-raising.", self, self.no_attempts
            )
            raise self.wrap_exception(self._exc)

    def _sync_attempts(self) -> Iterator[int]:
        # This is the same as _async_attempts, only synchronous
        delay = self.delay_min

        for attempt in range(1, self.no_attempts + 1):
            log.debug("[%r] Attempt %d of %d", self, attempt, self.no_attempts)

            self._exc = None

            yield attempt

            if not self._exc:
                break

            if attempt < self.no_attempts:
                log.debug("[%r] Retrying...", self)
                current_delay = delay + self.delay_add_fuzz * random()
                time.sleep(current_delay)
                delay = max(self.delay_max, delay * self.delay_backoff_factor)

        if self._exc:
            log.warning(
                "[%r] Number of attempts (%d) exhausted, re-raising.", self, self.no_attempts
            )
            raise self.wrap_exception(self._exc)

    @property
    def attempts(self) -> Union[AsyncIterator[int], Iterator[int]]:
        if self._is_async:
            return self._async_attempts()
        else:
            return self._sync_attempts()

    def exception_matches(self, exc: Exception) -> bool:
        """Check if exception should cause a retry.

        This method can be used to inspect the exception and re-raise it if it
        doesn't match certain criteria.
        """
        if not isinstance(exc, self.exceptions):
            log.debug(
                "[%r] Smoke tests failed (exception %s doesn’t match %r)",
                self,
                exc,
                self.exceptions,
            )
            return False
        return True

    def process_exception(self, exc: Exception) -> None:
        """Further process caught exception.

        This method needs to be called when an exception is caught.
        """
        exception_matches = self.exception_matches(exc)
        log.debug("[%r] Exception %r matches: %r", self, exc, exception_matches)
        if exception_matches:
            log.debug("[%r] Setting self._exc: %r", self, exc)
            self._exc = exc
        else:
            raise exc



"""ignore failed nodes in unique indexes

Revision ID: 36cf064c0481
Revises: ce2e575cb800
Create Date: 2023-07-27 11:29:37.405520
"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "36cf064c0481"
down_revision = "ce2e575cb800"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_index("active_hostname_index")
    op.create_index(
        "active_hostname_index",
        "nodes",
        ["hostname"],
        unique=True,
        postgresql_where=sa.text(
            "retired_at IS NULL AND state <> 'provisioning'::node_state_enum"
            + " AND state <> 'failed'::node_state_enum"
        ),
        sqlite_where=sa.text(
            "retired_at IS NULL AND state != 'provisioning' AND state != 'failed'"
        ),
    )
    op.drop_index("active_ipaddr_index")
    op.create_index(
        "active_ipaddr_index",
        "nodes",
        ["ipaddr"],
        unique=True,
        postgresql_where=sa.text(
            "retired_at IS NULL AND state <> 'provisioning'::node_state_enum"
            + " AND state <> 'failed'::node_state_enum"
        ),
        sqlite_where=sa.text(
            "retired_at IS NULL AND state != 'provisioning' AND state != 'failed'"
        ),
    )


def downgrade():
    op.drop_index("active_hostname_index")
    op.create_index(
        "active_hostname_index",
        "nodes",
        ["hostname"],
        unique=True,
        postgresql_where=sa.text("retired_at IS NULL AND state <> 'provisioning'::node_state_enum"),
        sqlite_where=sa.text("retired_at IS NULL AND state != 'provisioning'"),
    )
    op.drop_index("active_ipaddr_index")
    op.create_index(
        "active_ipaddr_index",
        "nodes",
        ["ipaddr"],
        unique=True,
        postgresql_where=sa.text("retired_at IS NULL AND state <> 'provisioning'::node_state_enum"),
        sqlite_where=sa.text("retired_at IS NULL AND state != 'provisioning'"),
    )



"""This is the pool controller."""

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy import func, select
from sqlalchemy.ext.asyncio import AsyncSession
from starlette.status import HTTP_404_NOT_FOUND, HTTP_422_UNPROCESSABLE_ENTITY

from ...api_models import PoolResult, PoolResultCollection
from ...database.model import Node
from ...nodes.pools import ConcreteNodePool
from ..database import req_db_async_session

router = APIRouter(prefix="/pools")


# http get http://localhost:8080/api/v1/pools
@router.get("", response_model=PoolResultCollection, tags=["pools"])
async def get_all_pools(db_async_session: AsyncSession = Depends(req_db_async_session)):
    """Return all pools."""
    pools = [
        {"name": pool.name, "fill-level": pool["fill-level"]}
        for pool in ConcreteNodePool.iter_pools()
        if "fill-level" in pool
    ]

    return {"action": "get", "pools": pools}


# http get http://localhost:8080/api/v1/pool/name-of-the-pool
@router.get("/{name}", response_model=PoolResult, tags=["pools"])
async def get_pool(name: str, db_async_session: AsyncSession = Depends(req_db_async_session)):
    """Return the pool with the specified **NAME**."""
    pool = ConcreteNodePool.known_pools.get(name)

    if pool is None:
        raise HTTPException(HTTP_404_NOT_FOUND)

    if "fill-level" not in pool:
        raise HTTPException(HTTP_422_UNPROCESSABLE_ENTITY)

    pool_result = {
        "name": name,
        "fill-level": pool["fill-level"],
        "levels": {
            "provisioning": 0,
            "ready": 0,
            "contextualizing": 0,
            "deployed": 0,
            "deprovisioning": 0,
        },
    }

    for state, quantity in await db_async_session.execute(
        select(Node.state, func.count(Node.state))
        .filter(Node.active == True, Node.pool == name)  # noqa: E712
        .group_by(Node.state)
    ):
        pool_result["levels"][state.name] = quantity

    return {"action": "get", "pool": pool_result}



import logging
import logging.config
import sys
from datetime import timedelta
from pathlib import Path
from typing import List, Optional, Tuple, Union

import click
from pydantic import TypeAdapter, ValidationError

try:
    import uvicorn
except ImportError:  # pragma: no cover
    uvicorn = None
import xdg.BaseDirectory
import yaml

try:
    from . import admin
except ImportError:  # pragma: no cover
    admin = None
try:
    from . import database
except ImportError:  # pragma: no cover
    database = None
try:
    from . import shell
except ImportError:  # pragma: no cover
    shell = None

from .app.logging import RequestIdFilter

try:
    from .client import DuffyClient, DuffyFormatter
except ImportError:  # pragma: no cover
    DuffyClient = DuffyFormatter = None
from .configuration import config, read_configuration

try:
    from .database.migrations.main import alembic_migration
except ImportError:  # pragma: no cover
    alembic_migration = None
try:
    from .database.setup import setup_db_schema, setup_db_test_data
except ImportError:  # pragma: no cover
    setup_db_schema = setup_db_test_data = None
from .exceptions import DuffyConfigurationError
from .misc import ConfigTimeDelta

try:
    from .tasks import start_worker
except ImportError:  # pragma: no cover
    start_worker = None
from .util import UNSET, SentinelType
from .version import __version__

DEFAULT_CONFIG_LOCATIONS = ("/etc/duffy", f"{xdg.BaseDirectory.xdg_config_home}/duffy")
DEFAULT_CONFIG_PATHS = tuple(Path(loc) for loc in DEFAULT_CONFIG_LOCATIONS)
LOGGING_FORMAT = "[%(request_id)s] %(levelname)s %(message)s"

log = logging.getLogger(__name__)


# Custom Click parameter types


class IntOrNoneType(click.ParamType):
    name = "int_or_none"

    def convert(self, value, param, ctx):
        if value is UNSET or isinstance(value, int):
            return value

        try:
            if isinstance(value, str) and value.lower() in ("none", "null"):
                return None

            return int(value)
        except ValueError:
            self.fail(f"{value!r} is not a valid integer", param, ctx)


INT_OR_NONE = IntOrNoneType()


class IntervalOrNoneType(click.ParamType):
    name = "interval_or_none"

    def convert(self, value, param, ctx):
        if value is UNSET:
            return value

        try:
            if isinstance(value, str) and value.lower() in ("none", "null"):
                return None

            return TypeAdapter(ConfigTimeDelta).validate_python(value)
        except ValidationError as exc:
            errors = exc.errors()
            self.fail(
                str(errors[0]["ctx"]["error"]) if errors else f"Can't convert value {value!r}",
                param,
                ctx,
            )


INTERVAL_OR_NONE = IntervalOrNoneType()


class NodesSpecType(click.ParamType):
    name = "nodes_spec"

    def convert(self, value, param, ctx):
        if value is None:
            return None

        nodes_spec = {}
        for item in value.split(","):
            key, value = item.split("=", 1)
            if key in nodes_spec:
                self.fail(f"Duplicate key: {key}")

            nodes_spec[key] = value

        if set(nodes_spec) != {"pool", "quantity"}:
            self.fail("Both `pool` and `quantity` must be set")

        return nodes_spec


NODES_SPEC = NodesSpecType()


# CLI groups and commands


@click.group(name="duffy")
@click.option(
    "-l",
    "--loglevel",
    "loglevel",
    type=(
        click.Choice(list(uvicorn.config.LOG_LEVELS.keys()), case_sensitive=False)
        if uvicorn
        else str
    ),
    help="Set the log level.",
    default=None,
)
@click.option(
    "config_paths",
    "-c",
    "--config",
    type=click.Path(exists=True),
    multiple=True,
    help=(
        "Read configuration from the specified YAML files or directories instead of the default"
        f" paths ({', '.join(DEFAULT_CONFIG_LOCATIONS)})"
    ),
    metavar="FILE_OR_DIR",
)
@click.version_option(version=__version__, prog_name="Duffy")
@click.pass_context
def cli(ctx: click.Context, loglevel: Optional[str], config_paths: Tuple[Path]):
    ctx.ensure_object(dict)
    ctx.obj["loglevel"] = loglevel

    logging.basicConfig(level=loglevel.upper() if isinstance(loglevel, str) else loglevel)

    if not config_paths:
        # Ignore non-existent default paths
        config_paths = tuple(path for path in DEFAULT_CONFIG_PATHS if path.exists())

    log.debug(f"Reading configuration from: {', '.join(str(p) for p in config_paths)}")

    read_configuration(*config_paths, clear=True, validate=True)


# Check & dump configuration


@cli.group(name="config")
def config_subcmd():
    """Check and dump configuration."""


@config_subcmd.command(name="check")
def config_check():
    """Validate configuration structure.

    This checks if configuration subkeys conform to the expected format.
    However, it doesn't check if the sub-keys necessary to run a certain
    subcommand exist.
    """
    if not config:
        click.echo("Configuration is empty.")
    else:
        click.echo(f"OK.\nValidated configuration subkeys: {', '.join(config)}")


@config_subcmd.command("dump")
def config_dump():
    """Dump merged configuration."""
    yaml.safe_dump(config, sys.stdout)


# Set up the database tables


@cli.command()
@click.option(
    "--test-data/--no-test-data", default=False, help="Initialized database with test data."
)
def setup_db(test_data):
    """Create tables from the database model."""
    if not (database and setup_db_schema and setup_db_test_data):
        raise click.ClickException(
            "Please install the duffy[database] extra (and optionally duffy[postgresql] or"
            " duffy[sqlite]) for this command"
        )

    try:
        setup_db_schema()
        if test_data:
            database.init_model()
            setup_db_test_data()
    except DuffyConfigurationError as exc:
        log.error("Configuration key missing or wrong: %s", exc.args[0])
        sys.exit(1)


# Handle database migrations


@cli.group()
def migration():
    """Handle database migrations."""
    if not alembic_migration:
        raise click.ClickException(
            "Please install the duffy[database] extra (and optionally duffy[postgresql] or"
            " duffy[sqlite]) for this command"
        )


@migration.command("create")
@click.option(
    "--autogenerate/--no-autogenerate",
    default=False,
    help="Autogenerate migration script skeleton (needs to be reviewed/edited).",
)
@click.argument("comment", nargs=-1, required=True)
def migration_create(autogenerate, comment):
    """Create a new database schema migration."""
    alembic_migration.create(comment=" ".join(comment), autogenerate=autogenerate)


@migration.command("db-version")
def migration_db_version():
    """Show the current version of the database schema."""
    alembic_migration.db_version()


@migration.command("upgrade")
@click.argument("version", default="head")
def migration_upgrade(version):
    """Upgrade the database schema."""
    alembic_migration.upgrade(version)


@migration.command("downgrade")
@click.argument("version", default="-1")
def migration_downgrade(version):
    """Downgrade the database schema."""
    alembic_migration.downgrade(version)


# Interactive development/debugging shell


@cli.command(name="dev-shell")
@click.option(
    "-t",
    "--shell-type",
    type=click.Choice(shell.get_available_shells(), case_sensitive=False) if shell else str,
    help="Type of interactive shell to use.",
    default=None,
)
def dev_shell(shell_type: str):
    """Run an interactive shell."""
    try:
        database.init_model()
    except DuffyConfigurationError as exc:
        log.error("Configuration key missing or wrong: %s", exc.args[0])
        sys.exit(1)

    shell.embed_shell(shell_type=shell_type)


# Run the backend task worker


@cli.command(context_settings={"ignore_unknown_options": True, "help_option_names": []})
@click.argument("worker_args", nargs=-1, type=click.UNPROCESSED)
def worker(worker_args: Tuple[str]):
    """Start a Celery worker to process backend tasks."""
    if not start_worker:
        raise click.ClickException("Please install the duffy[tasks] extra for this command")

    start_worker(worker_args=worker_args)


# Run the web app


@cli.command()
@click.option(
    "--reload/--no-reload", default=False, help="Automatically reload if the code is changed."
)
@click.option("-H", "--host", default=None, help="Set the host address to listen on.")
@click.option(
    "-p",
    "--port",
    type=click.IntRange(1, 65535, clamp=True),
    default=None,
    help="Set the port value.",
)
@click.pass_obj
def serve(obj, reload, host, port):
    """Run the Duffy web application server.

    Duffy is the middle layer running ci.centos.org that manages the
    provisioning, maintenance, teardown and rebuild of the Nodes
    (physical hardware and virtual machines) that are used to run
    the tests in the CI Cluster.
    """
    if not uvicorn:
        raise click.ClickException(
            "Please install the duffy[app] extra (and optionally duffy[postgresql] or"
            " duffy[sqlite]) for this command"
        )

    config_app = config.get("app", {})
    if host is None:
        host = config_app.get("host", "127.0.0.1")
    if port is None:
        port = config_app.get("port", 8080)
    loglevel = obj["loglevel"]
    if loglevel is None:
        loglevel = config_app.get("loglevel", "info")
    numeric_loglevel = uvicorn.config.LOG_LEVELS[loglevel.lower()]

    handler = logging.StreamHandler()
    handler.addFilter(RequestIdFilter())
    logging.basicConfig(
        force=True,
        level=numeric_loglevel,
        format=LOGGING_FORMAT,
        handlers=[handler],
    )

    config_logging = config_app.get("logging")
    if config_logging:
        logging.config.dictConfig(config_logging)

    # Report for duty
    print(" * Starting Duffy...")
    print(f" * Host address : {host}")
    print(f" * Port number  : {port}")
    print(f" * Log level    : {loglevel}")
    print(f" * Serving API docs on http://{host}:{port}/docs")

    # Start the show
    uvicorn.run(
        "duffy.app.main:app",
        host=host,
        port=port,
        log_level=numeric_loglevel,
        reload=reload,
        log_config=None,
    )


# Run the web app - Duffy Metaclient for Legacy Support


@cli.command()
@click.option(
    "--reload/--no-reload", default=False, help="Automatically reload if the code is changed."
)
@click.option("-H", "--host", default=None, help="Set the host address to listen on.")
@click.option(
    "-p",
    "--port",
    type=click.IntRange(1, 65535, clamp=True),
    default=None,
    help="Set the port value.",
)
@click.option(
    "-D",
    "--dest",
    default=None,
    help="Set the destination address of Duffy deployment.",
)
@click.pass_obj
def serve_legacy(obj, reload, host, port, dest):
    """Serve the Duffy Metaclient for Legacy Support app.

    Duffy is the middle layer running ci.centos.org that manages the
    provisioning, maintenance and teardown / rebuild of the Nodes
    (physical hardware for now, VMs coming soon) that are used to run
    the tests in the CI Cluster.

    This metaclient exposes older endpoints for legacy support and
    connects them to the path operations introduced by the newer version
    of the Duffy endpoint, until the support for the older endpoints is
    deprecated.
    """
    if not uvicorn:
        raise click.ClickException("Please install the duffy[legacy] extra for this command")

    config_metaclient = config.get("metaclient", {})
    if host is None:
        host = config_metaclient.get("host", "127.0.0.1")
    if port is None:
        port = config_metaclient.get("port", 9090)
    if dest is None:
        dest = config_metaclient.get("dest", "http://127.0.0.1:8080")
    loglevel = obj["loglevel"]
    if loglevel is None:
        loglevel = config_metaclient.get("loglevel", "info")
    numeric_loglevel = uvicorn.config.LOG_LEVELS[loglevel.lower()]

    handler = logging.StreamHandler()
    handler.addFilter(RequestIdFilter())
    logging.basicConfig(
        force=True,
        level=numeric_loglevel,
        format=LOGGING_FORMAT,
        handlers=[handler],
    )

    config_logging = config_metaclient.get("logging")
    if config_logging:
        logging.config.dictConfig(config_logging)

    # Report for duty
    print(" * Starting Duffy Metaclient for Legacy Support...")
    print(f" * Host address        : {host}")
    print(f" * Port number         : {port}")
    print(f" * Destination address : {dest}")
    print(f" * Log level           : {loglevel}")
    print(f" * Serving API docs on http://{host}:{port}/docs")

    # Start the show
    uvicorn.run(
        "duffy.legacy.main:app",
        host=host,
        port=port,
        log_level=numeric_loglevel,
        reload=reload,
        log_config=None,
    )


# Commands for administrative tasks


class FakeAPITenant:
    is_admin = True


@cli.group("admin")
def admin_group():
    """Administrate Duffy tenants."""
    if not admin:
        raise click.ClickException(
            "Please install the duffy[admin] extra (and optionally duffy[postgresql] or"
            " duffy[sqlite]) for this command"
        )


@admin_group.command("list-tenants")
@click.option("--quiet/--no-quiet", default=False, help="Show only tenant information.")
@click.option("--active/--all", default=True, help="Whether to list retired tenants.")
def admin_list_tenants(quiet, active):
    """List tenants."""
    admin_ctx = admin.AdminContext.create_for_cli()
    result = admin_ctx.list_tenants()
    if "error" in result:
        click.echo(
            f"ERROR: couldn't list tenants\nERROR DETAIL: {result['error']['detail']}", err=True
        )
        sys.exit(1)
    else:
        prefix = "OK: " if not quiet else ""
        for tenant in sorted(result["tenants"], key=lambda t: t.name):
            if active and not tenant.active:
                continue
            click.echo(f"{prefix}{tenant.name}")


@admin_group.command("show-tenant")
@click.argument("name")
def admin_show_tenant(name: str):
    """Show a tenant."""
    admin_ctx = admin.AdminContext.create_for_cli()
    result = admin_ctx.show_tenant(name)
    if "error" in result:
        click.echo(f"ERROR: {name}\nERROR DETAIL: {result['error']['detail']}", err=True)
        sys.exit(1)
    else:
        tenant = result["tenant"]
        click.echo(
            f"OK: {name}: id={tenant.id} node_quota={tenant.node_quota}"
            f" effective_node_quota={tenant.effective_node_quota}"
            f" session_lifetime={tenant.session_lifetime}"
            f" effective_session_lifetime={tenant.effective_session_lifetime}"
            f" session_lifetime_max={tenant.session_lifetime_max}"
            f" effective_session_lifetime_max={tenant.effective_session_lifetime_max}"
            f" active={tenant.active} created_at={tenant.created_at} retired_at={tenant.retired_at}"
        )


@admin_group.command("create-tenant")
@click.option(
    "--is-admin/--no-is-admin",
    default=False,
    help="If the tenant should have administrative rights",
)
@click.option(
    "--node-quota",
    type=INT_OR_NONE,
    default=None,
    help="How many nodes the tenant can use at a time (optional, will use default if unset).",
)
@click.option(
    "--session-lifetime",
    type=INTERVAL_OR_NONE,
    default=None,
    help="The initial session lifetime for this tenant.",
)
@click.option(
    "--session-lifetime-max",
    type=INTERVAL_OR_NONE,
    default=None,
    help="The maximum session lifetime for this tenant.",
)
@click.argument("name")
@click.argument("ssh_key")
def admin_create_tenant(
    name: str,
    ssh_key: str,
    is_admin: bool,
    node_quota: Optional[int],
    session_lifetime: Optional[timedelta],
    session_lifetime_max: Optional[timedelta],
):
    """Create a new tenant."""
    admin_ctx = admin.AdminContext.create_for_cli()
    result = admin_ctx.create_tenant(
        name=name,
        ssh_key=ssh_key,
        node_quota=node_quota,
        session_lifetime=session_lifetime,
        session_lifetime_max=session_lifetime_max,
        is_admin=is_admin,
    )
    if "error" in result:
        click.echo(f"ERROR: {name}\nERROR DETAIL: {result['error']['detail']}", err=True)
        sys.exit(1)
    else:
        click.echo(f"OK: {name}: {result['tenant'].api_key}")


@admin_group.command("retire-tenant")
@click.option("--retire/--unretire", default=True, help="Whether to retire or unretire a tenant.")
@click.argument("name")
def admin_retire_tenant(name: str, retire: bool = True):
    """Retire or unretire a tenant."""
    admin_ctx = admin.AdminContext.create_for_cli()
    result = admin_ctx.retire_unretire_tenant(name, retire=retire)
    if "error" in result:
        click.echo(f"ERROR: {name}\nERROR DETAIL: {result['error']['detail']}", err=True)
        sys.exit(1)
    else:
        click.echo(f"OK: {name}: active={result['tenant'].active}")


@admin_group.command("update-tenant")
@click.option("--ssh-key", help="New SSH key for the tenant.")
@click.option(
    "--api-key", help="Either a new API key (UUID) for the tenant or 'reset' to set automatically."
)
@click.option(
    "--node-quota",
    type=INT_OR_NONE,
    default=UNSET,
    help="How many nodes the tenant can use at a time (optional, will use default if unset).",
)
@click.option(
    "--session-lifetime",
    type=INTERVAL_OR_NONE,
    default=UNSET,
    help="The initial session lifetime for this tenant.",
)
@click.option(
    "--session-lifetime-max",
    type=INTERVAL_OR_NONE,
    default=UNSET,
    help="The maximum session lifetime for this tenant.",
)
@click.argument("name")
def admin_update_tenant(
    name: str,
    node_quota: Optional[Union[int, SentinelType]],
    session_lifetime: Optional[Union[timedelta, SentinelType]],
    session_lifetime_max: Optional[Union[timedelta, SentinelType]],
    ssh_key: str = None,
    api_key: str = None,
):
    """Update a tenant."""
    if (
        not ssh_key
        and not api_key
        and node_quota is UNSET
        and session_lifetime is UNSET
        and session_lifetime_max is UNSET
    ):
        click.echo(
            "ERROR: Either --ssh-key, --api-key, --node-quota, --session-lifetime or"
            " --session-lifetime-max must be set.",
            err=True,
        )
        sys.exit(1)
    admin_ctx = admin.AdminContext.create_for_cli()
    kwargs = {"name": name, "ssh_key": ssh_key, "api_key": api_key}
    if node_quota is not UNSET:
        kwargs["node_quota"] = node_quota
    if session_lifetime is not UNSET:
        kwargs["session_lifetime"] = session_lifetime
    if session_lifetime_max is not UNSET:
        kwargs["session_lifetime_max"] = session_lifetime_max
    result = admin_ctx.update_tenant(**kwargs)
    if "error" in result:
        click.echo(f"ERROR: {name}\nERROR DETAIL: {result['error']['detail']}", err=True)
        sys.exit(1)
    else:
        tenant = result["tenant"]
        click.echo(
            f"OK: {name}: ssh_key={tenant.ssh_key} api_key={tenant.api_key}"
            f" node_quota={tenant.node_quota} effective_node_quota={tenant.effective_node_quota}"
            f" session_lifetime={tenant.session_lifetime}"
            f" effective_session_lifetime={tenant.effective_session_lifetime}"
            f" session_lifetime_max={tenant.session_lifetime_max}"
            f" effective_session_lifetime_max={tenant.effective_session_lifetime_max}"
        )


@cli.group()
@click.option("--url", help="The base URL of the Duffy API.")
@click.option("--auth-name", help="The tenant name to authenticate with the Duffy API.")
@click.option("--auth-key", help="The tenant key to authenticate with the Duffy API.")
@click.option(
    "--format",
    type=click.Choice(
        DuffyFormatter._subclasses_for_format.keys()
        if DuffyFormatter
        else ("json", "yaml", "flat"),
        case_sensitive=False,
    ),
    default="json",
    help="Format with which to print results.",
)
@click.pass_context
def client(
    ctx: click.Context,
    url: Optional[str],
    auth_name: Optional[str],
    auth_key: Optional[str],
    format: str,
):
    """Command line client for the Duffy API."""
    if not (DuffyClient and DuffyFormatter):
        raise click.ClickException("Please install the duffy[client] extra for this command")

    ctx.ensure_object(dict)
    ctx.obj["client"] = DuffyClient(url=url, auth_name=auth_name, auth_key=auth_key)
    ctx.obj["formatter"] = DuffyFormatter.new_for_format(format)


@client.command("list-sessions")
@click.pass_obj
def client_list_sessions(obj):
    """Query active sessions for this tenant on the Duffy API."""
    result = obj["client"].list_sessions()
    formatted_result = obj["formatter"].format(result)
    # Only print newline if formatted_result isn't empty.
    click.echo(formatted_result, nl=formatted_result)


@client.command("show-session")
@click.argument("session_id", type=int)
@click.pass_obj
def client_show_session(obj, session_id: int):
    """Show one session identified by its id on the Duffy API."""
    result = obj["client"].show_session(session_id)
    click.echo(obj["formatter"].format(result))


@client.command("request-session")
@click.argument(
    "nodes_specs",
    type=NODES_SPEC,
    nargs=-1,
    required=True,
    metavar="pool=<pool>,quantity=<quantity> [...]",
)
@click.pass_obj
def client_request_session(obj: dict, nodes_specs: List[str]):
    """Request a session with nodes from the Duffy API."""
    result = obj["client"].request_session(nodes_specs)
    click.echo(obj["formatter"].format(result))


@client.command("retire-session")
@click.argument("session_id", type=int)
@click.pass_obj
def client_retire_session(obj: dict, session_id: int):
    """Retire an active Duffy session."""
    result = obj["client"].retire_session(session_id)
    click.echo(obj["formatter"].format(result))


@client.command("list-pools")
@click.pass_obj
def client_list_pools(obj: dict):
    """List configured Duffy node pools."""
    result = obj["client"].list_pools()
    formatted_result = obj["formatter"].format(result)
    # Only print newline if formatted_result isn't empty.
    click.echo(formatted_result, nl=formatted_result)


@client.command("show-pool")
@click.argument("name")
@click.pass_obj
def client_show_pool(obj: dict, name: str):
    """Show information about a Duffy node pool."""
    result = obj["client"].show_pool(name)
    click.echo(obj["formatter"].format(result))



from abc import ABC
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, IPvAnyAddress

try:
    from ..database.types import NodeState
except ImportError:  # pragma: no cover
    NodeState = str
from .common import APIResult, CreatableMixin, RetirableMixin

# abstract node


class NodeBase(BaseModel, ABC):
    hostname: Optional[str] = None
    ipaddr: Optional[IPvAnyAddress] = None
    comment: Optional[str] = None

    pool: Optional[str] = None
    reusable: bool
    data: Dict[str, Any]
    model_config = ConfigDict(from_attributes=True)


class NodeCreateModel(NodeBase):
    hostname: str
    ipaddr: IPvAnyAddress
    reusable: bool = True
    data: Dict[str, Any] = {}


class NodeModel(NodeBase, CreatableMixin, RetirableMixin):
    id: int
    state: NodeState


# API results


class NodeResult(APIResult):
    node: NodeModel


class NodeResultCollection(APIResult):
    nodes: List[NodeModel]



from contextvars import ContextVar
from typing import Optional
from uuid import uuid4

from starlette.datastructures import MutableHeaders
from starlette.types import ASGIApp, Message, Receive, Scope, Send

request_id_ctxvar: ContextVar[Optional[str]] = ContextVar("request-id", default=None)


class RequestIdMiddleware:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        request_id = str(uuid4())
        request_id_ctxvar.set(request_id)

        if scope["type"] != "http":  # pragma: no cover
            await self.app(scope, receive, send)
            return

        async def send_with_extra_header(message: Message) -> None:
            if message["type"] == "http.response.start":
                headers = MutableHeaders(scope=message)
                headers.append("X-Request-Id", request_id)

            await send(message)

        await self.app(scope, receive, send_with_extra_header)



from abc import ABC
from typing import List, Union

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Annotated

from .common import APIResult

# pool model


class PoolLevelsModel(BaseModel):
    provisioning: Annotated[int, Field(ge=0)]
    ready: Annotated[int, Field(ge=0)]
    contextualizing: Annotated[int, Field(ge=0)]
    deployed: Annotated[int, Field(ge=0)]
    deprovisioning: Annotated[int, Field(ge=0)]


class PoolBase(BaseModel, ABC):
    name: str
    fill_level: Annotated[int, Field(ge=0)] = Field(alias="fill-level")
    model_config = ConfigDict(extra="forbid")


class PoolConciseModel(PoolBase):
    pass


class PoolVerboseModel(PoolConciseModel):
    levels: PoolLevelsModel


PoolModel = Union[PoolConciseModel, PoolVerboseModel]


# API results


class PoolResult(APIResult):
    pool: PoolModel


class PoolResultCollection(APIResult):
    pools: List[PoolModel]



import logging
from typing import Tuple, Union

try:
    import asyncpg
except ImportError:  # pragma: no cover
    asyncpg = None
from sqlalchemy.exc import DBAPIError

from ..configuration import config_get
from ..util import RetryContext

log = logging.getLogger(__name__)


class ConfigRetryContext(RetryContext):
    def __init__(self, *, exceptions: Union[Exception, Tuple[Exception]] = None, **kwargs):
        for key in (
            "no-attempts",
            "delay-min",
            "delay-max",
            "delay-backoff-factor",
            "delay-add-fuzz",
        ):
            key_ = key.replace("-", "_")
            if key_ not in kwargs:
                kwargs[key_] = config_get(f"app.retries.{key}", f"defaults.retries.{key}")
        super().__init__(exceptions=exceptions, **kwargs)


class SerializationErrorRetryContext(ConfigRetryContext):
    exceptions = DBAPIError

    def exception_matches(self, exc):
        log.debug("[%r] Matching %r...", self, exc)
        if not super().exception_matches(exc):
            return False

        result = getattr(exc.orig, "pgcode", None) == asyncpg.SerializationError.sqlstate
        log.debug("[%r] Match result: %r", self, result)

        return result



import re
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union
from uuid import UUID

from pydantic import (
    AnyHttpUrl,
    AnyUrl,
    BaseModel,
    ConfigDict,
    Field,
    RedisDsn,
    UrlConstraints,
    field_validator,
)
from typing_extensions import Annotated

from ..misc import ConfigTimeDelta

# enums


class LogLevel(str, Enum):
    trace = "trace"
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"
    critical = "critical"


class MechanismType(str, Enum):
    ansible = "ansible"


# Pydantic models


class ConfigBaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid")


class CeleryModel(BaseModel):
    # This is intentionally not a subclass of ConfigBaseModel, it is passed on to Celery, i.e. can
    # contain arbitrarily named fields.
    broker_url: AnyUrl
    result_backend: AnyUrl


class LockingModel(ConfigBaseModel):
    url: RedisDsn


class PeriodicTaskModel(ConfigBaseModel):
    interval: ConfigTimeDelta


class TasksModel(ConfigBaseModel):
    celery: CeleryModel
    locking: LockingModel
    periodic: Optional[Dict[str, PeriodicTaskModel]] = None


class SQLAlchemyModel(BaseModel):
    # This is intentionally not a subclass of ConfigBaseModel, it is passed on to SQLAlchemy's
    # create_engine()/create_async_engine(), i.e. can contain arbitrarily named fields.
    sync_url: Annotated[AnyUrl, UrlConstraints(host_required=False)]
    async_url: Annotated[AnyUrl, UrlConstraints(host_required=False)]


class DatabaseModel(ConfigBaseModel):
    sqlalchemy: SQLAlchemyModel


class RetriesModel(ConfigBaseModel):
    no_attempts: Optional[Annotated[int, Field(ge=1)]] = Field(alias="no-attempts", default=None)
    delay_min: Optional[Union[Annotated[int, Field(ge=0)], Annotated[float, Field(ge=0)]]] = Field(
        alias="delay-min", default=None
    )
    delay_max: Optional[Union[Annotated[int, Field(ge=0)], Annotated[float, Field(ge=0)]]] = Field(
        alias="delay-max", default=None
    )
    delay_backoff_factor: Optional[
        Union[Annotated[int, Field(ge=1)], Annotated[float, Field(ge=1)]]
    ] = Field(alias="delay-backoff-factor", default=None)
    delay_add_fuzz: Optional[
        Union[Annotated[int, Field(ge=0)], Annotated[float, Field(ge=0)]]
    ] = Field(alias="delay-add-fuzz", default=None)


class DefaultsModel(ConfigBaseModel):
    session_lifetime: ConfigTimeDelta = Field(alias="session-lifetime")
    session_lifetime_max: ConfigTimeDelta = Field(alias="session-lifetime-max")
    node_quota: Annotated[int, Field(gt=0)] = Field(alias="node-quota")
    retries: Optional[RetriesModel] = None


class AnsibleMechanismPlaybookModel(ConfigBaseModel):
    extra_vars: Optional[Dict[str, Any]] = Field(alias="extra-vars", default=None)
    playbook: Path


class AnsibleMechanismModel(ConfigBaseModel):
    topdir: Optional[Path] = None
    extra_vars: Optional[Dict[str, Any]] = Field(alias="extra-vars", default=None)
    provision: Optional[AnsibleMechanismPlaybookModel] = None
    deprovision: Optional[AnsibleMechanismPlaybookModel] = None


class MechanismModel(ConfigBaseModel):
    type_: Optional[MechanismType] = Field(alias="type", default=None)
    ansible: Optional[AnsibleMechanismModel] = None


class NodePoolsModel(ConfigBaseModel):
    extends: Optional[str] = None
    mechanism: Optional[MechanismModel] = None
    fill_level: Optional[Annotated[int, Field(gt=0)]] = Field(alias="fill-level", default=None)
    reuse_nodes: Optional[Union[Dict[str, Union[int, str]], Literal[False]]] = Field(
        alias="reuse-nodes", default=None
    )
    run_parallel: Optional[bool] = Field(alias="run-parallel", default=None)
    model_config = ConfigDict(extra="allow")


class NodePoolsRootModel(ConfigBaseModel):
    abstract: Optional[Dict[str, NodePoolsModel]] = None
    concrete: Dict[str, NodePoolsModel]


class LoggingModel(ConfigBaseModel):
    version: Literal[1]
    model_config = ConfigDict(extra="allow")


class ClientAuthModel(ConfigBaseModel):
    name: str
    key: UUID


class ClientModel(ConfigBaseModel):
    url: AnyHttpUrl
    auth: ClientAuthModel


class AppModel(ConfigBaseModel):
    loglevel: Optional[LogLevel] = None
    host: Optional[str] = None
    port: Optional[Annotated[int, Field(gt=0, lt=65536)]] = None
    logging: Optional[LoggingModel] = None
    retries: Optional[RetriesModel] = None


class LegacyPoolMapModel(ConfigBaseModel):
    pool: str
    ver: Optional[str] = None
    arch: Optional[str] = None
    flavor: Optional[str] = None

    @field_validator("ver", "arch", "flavor")
    @classmethod
    def detect_regex(cls, v: str):
        if v.startswith("^") and v.endswith("$"):
            return re.compile(v)
        return v


class LegacyModel(ConfigBaseModel):
    host: Optional[str] = None
    port: Optional[Annotated[int, Field(gt=0, lt=65536)]] = None
    dest: Optional[AnyHttpUrl] = None
    loglevel: Optional[LogLevel] = None
    logging: Optional[LoggingModel] = None
    usermap: Dict[str, str]
    poolmap: List[LegacyPoolMapModel]
    mangle_hostname: Optional[str] = None


class ConfigModel(ConfigBaseModel):
    client: Optional[ClientModel] = None
    app: Optional[AppModel] = None
    tasks: Optional[TasksModel] = None
    database: Optional[DatabaseModel] = None
    defaults: Optional[DefaultsModel] = None
    metaclient: Optional[LegacyModel] = None
    nodepools: Optional[NodePoolsRootModel] = None
